(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory NLPolyOrder imports TermOrder NLPolynomial begin

types ('f,'a)poly_inter = "'f \<Rightarrow> nat \<Rightarrow> (nat,'a)poly"

fun eval_term :: "('f,'a :: poly_carrier)poly_inter \<Rightarrow> ('f,'v)term \<Rightarrow> ('v,'a)poly"
where "eval_term _ (Var x) = [([(x,1)],1)]"
  |   "eval_term I (Fun f ts) = (let ps = map (eval_term I) ts; n = length ts in poly_subst (\<lambda> i. if i < n then ps ! i else zero_poly) (I f n))"


locale poly_order = poly_order_carrier +
  fixes I :: "('f,'a)poly_inter"
    and type :: 'v
    and \<pi> :: "'f af"
  assumes pos_I: "\<And> f n. I f n \<ge>p zero_poly"
  and mono_I: "\<And> f n. poly_weak_mono type (I f n)" 
  and pi: "\<And> f n i. \<not> (\<pi> f n i) \<Longrightarrow> \<not> (poly_var (I f n) i)"
begin

lemma eval_term_pos: fixes t :: "('f,'v)term" shows "eval_term I t \<ge>p zero_poly"
unfolding poly_ge_def zero_poly_def
proof (intro impI allI, simp)
  fix \<alpha> :: "('v,'a)assign"
  assume pos: "pos_assign \<alpha>"
  show "eval_poly \<alpha> (eval_term I t) \<ge> 0"
  proof (induct t, simp add: pos[unfolded pos_assign_def])
    case (Fun f ts)
    show ?case 
    proof (simp add: Let_def poly_subst, rule pos_assign_poly[OF _ pos_I], unfold pos_assign_def, intro allI)
      fix i
      show "eval_poly \<alpha> (if i < length ts then map (eval_term I) ts ! i else zero_poly) \<ge> 0"
      proof (cases "i < length ts")
        case False thus ?thesis unfolding zero_poly_def by (simp add: ge_refl)
      next
        case True
        hence "ts ! i \<in> set ts" by auto
        with Fun have "eval_poly \<alpha> (eval_term I (ts ! i)) \<ge> 0" by auto
        with True show ?thesis by (simp)
      qed
    qed
  qed
qed

definition inter_s :: "('f,'v)trs"
  where "inter_s \<equiv> {(s,t). (eval_term I s >p eval_term I t)}"

definition inter_ns :: "('f,'v)trs"
  where "inter_ns \<equiv> {(s,t). eval_term I s \<ge>p eval_term I t}"

lemma inter_stable: 
  shows "eval_poly \<alpha> (eval_term I (t \<cdot> \<sigma>)) = eval_poly (\<lambda> x. eval_poly \<alpha> (eval_term I (get_subst \<sigma> x))) (eval_term I t)"
proof (cases \<sigma>)
  case (Subst sig)
  have "eval_poly \<alpha> (eval_term I (t \<cdot> Subst sig)) = eval_poly (\<lambda> x. eval_poly \<alpha> (eval_term I (sig x))) (eval_term I t)" (is "_ = eval_poly ?ass _")
  proof (induct t)
    case (Var x)
    thus ?case by simp
  next
    case (Fun f ts)
    let ?ts = "map (\<lambda> s. s \<cdot> Subst sig) ts"
    show ?case
    proof (simp add: Let_def poly_subst, rule fun_cong, rule arg_cong[where f = eval_poly], rule ext, unfold zero_poly_def)
      fix i
      show "eval_poly \<alpha> (if i < length ts then map (eval_term I \<circ> (\<lambda> t. t \<cdot> Subst sig)) ts ! i else []) = 
            eval_poly ?ass (if i < length ts then map (eval_term I) ts ! i else [])"
      proof (cases "i < length ts")
        case False thus ?thesis by simp
      next
        case True
        hence "ts ! i \<in> set ts" by auto
        from Fun[OF this] True show ?thesis by simp
      qed
    qed
  qed
  with Subst show ?thesis by simp
qed

end

sublocale poly_order \<subseteq> af_redpair inter_s inter_ns \<pi>
proof
  show "subst_closed inter_s"
    unfolding subst_closed_def
  proof
    fix ss ts :: "('f,'v)term"
    assume "(ss,ts) \<in> CS inter_s"
    thus "(ss,ts) \<in> inter_s"
    proof
      fix s t sig
      assume ss: "ss = s \<cdot> sig" and ts: "ts = t \<cdot> sig" and st: "(s,t) \<in> inter_s"
      thus ?thesis using eval_term_pos[unfolded poly_ge_def zero_poly_def] 
        by (force simp: inter_s_def poly_gt_def inter_stable pos_assign_def eval_term_pos)
    qed
  qed
next
  show "subst_closed inter_ns"
    unfolding subst_closed_def
  proof
    fix ss ts :: "('f,'v)term"
    assume "(ss,ts) \<in> CS inter_ns"
    thus "(ss,ts) \<in> inter_ns"
    proof
      fix s t sig
      assume ss: "ss = s \<cdot> sig" and ts: "ts = t \<cdot> sig" and st: "(s,t) \<in> inter_ns"
      thus ?thesis using eval_term_pos[unfolded poly_ge_def zero_poly_def] 
        by (force simp: inter_ns_def poly_ge_def inter_stable pos_assign_def eval_term_pos)
    qed
  qed
next
  show "SN inter_s"
  proof
    fix f
    assume steps: "\<forall> i. (f i, f (Suc i)) \<in> inter_s"
    obtain g where g: "\<And> i. g i = eval_term I (f i)" by auto
    from steps[unfolded inter_s_def] eval_term_pos
    have "\<And> i. (g i, g (Suc i)) \<in> poly_GT" 
      by (simp only: g, auto)
    with poly_GT_SN show False unfolding SN_defs by blast
  qed
next
  show "inter_ns O inter_s \<subseteq> inter_s"
    by (rule subsetI2, clarify, unfold inter_s_def inter_ns_def, simp add: poly_compat)
next
  show "one_step_ctxt_closed inter_ns"
    unfolding one_step_ctxt_closed_def inter_ns_def
  proof (intro allI impI, clarify)
    fix f bef and s t :: "('f,'v)term" and aft
    assume st: "eval_term I s \<ge>p eval_term I t"
    show "eval_term I (Fun f (bef @ s # aft)) \<ge>p eval_term I (Fun f (bef @ t # aft))"
    proof -
      obtain ss ts n where ss: "ss = bef @ s # aft" and ts: "ts = bef @ t # aft" and lss: "length ss = n" and lts: "length ts = n" by auto
      obtain p where p: "I f n = p" and mono: "poly_weak_mono type p" using mono_I by auto 
      { 
        fix i
        have "eval_term I ((bef @ s # aft) ! i) \<ge>p eval_term I ((bef @ t # aft) ! i)" 
        proof (cases "i < length bef")
          case True
          thus ?thesis by (auto simp: nth_append)
        next
          case False note oFalse = this
          show ?thesis
          proof (cases "i = length bef")
            case True
            thus ?thesis by (auto simp: nth_append st) 
          next
            case False
            with oFalse have "i > length bef" by auto
            hence "i - length bef > 0" by auto
            then obtain k where k: "i - length bef = Suc k" by (cases "i - length bef", auto)
            show ?thesis by (simp add: nth_append k)
          qed
        qed        
        hence "eval_term I (ss ! i) \<ge>p eval_term I (ts ! i)" by (simp add: ss ts)
      } note res = this
      from lts lss have l: "length ts = length ss" by auto
      show ?thesis by  (simp only: ss[symmetric] ts[symmetric], simp add: Let_def lss lts p, rule
         mono[unfolded poly_weak_mono_def, THEN spec, THEN spec, THEN mp], intro allI conjI, auto simp: lts[symmetric] eval_term_pos, simp add: l res)
    qed
  qed
next
  show "af_compatible \<pi> inter_ns"
    unfolding af_compatible_def
  proof (intro allI, clarify)
    fix f bef s t aft
    assume ns: "(Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<notin> inter_ns"
    let ?s = "bef @ s # aft"
    let ?t = "bef @ t # aft"
    let ?n = "Suc (length bef + length aft)"
    let ?i = "length bef"
    show "\<pi> f ?n ?i"
    proof (rule ccontr)
      assume "\<not> ?thesis"
      with pi have pv: "\<not> poly_var (I f ?n) ?i" by simp
      have "(Fun f ?s, Fun f ?t) \<in> inter_ns" unfolding inter_ns_def poly_ge_def
      proof (clarify)
        fix \<alpha>
        have "eval_poly \<alpha> (eval_term I (Fun f ?s)) = eval_poly \<alpha> (eval_term I (Fun f ?t))" 
        proof (simp, rule arg_cong[where f = "eval_poly \<alpha>"], rule poly_var[OF pv])
          fix j
          assume "?i \<noteq> j"
          thus "(if j < ?n then (map (eval_term I) bef @ eval_term I s # map (eval_term I) aft) ! j else zero_poly) = 
            (if j < ?n then (map (eval_term I) bef @ eval_term I t # map (eval_term I) aft) ! j else zero_poly)"
            by (cases "j \<ge> ?n", simp, cases "j < ?i", simp add: nth_append, simp, cases "j \<ge> ?i", simp add: nth_append, cases "j - ?i", simp, simp add: nth_append, simp)
        qed
        thus "eval_poly \<alpha> (eval_term I (Fun f ?s)) \<ge> eval_poly \<alpha> (eval_term I (Fun f ?t))" using ge_refl by simp
      qed
      with ns show False ..
    qed
  qed
qed

locale mono_poly_order = poly_order +
  assumes strict_mono_I: "\<And> f n i. i < n \<Longrightarrow> poly_strict_mono type (I f n) i"
begin

lemma inter_s_mono: "one_step_ctxt_closed inter_s"
unfolding one_step_ctxt_closed_def
proof (intro allI impI)
  fix f bef and s t :: "('b,'c)term" and aft 
  assume st: "(s,t) \<in> inter_s"
  show "(Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> inter_s" (is "(Fun f ?s, Fun f ?t) \<in> _")
    unfolding inter_s_def poly_gt_def
  proof (clarify)
    fix \<alpha> :: "('c,'a)assign"
    assume pos: "pos_assign \<alpha>"
    let ?n = "Suc (length bef + length aft)"
    let ?i = "length bef"
    from strict_mono_I[of ?i ?n] have mono: "poly_strict_mono type (I f ?n) ?i" by simp
    let ?exp = "\<lambda> w s.  (if w < Suc (length bef + length aft) then (map (eval_term I) bef @ eval_term I s # map (eval_term I) aft) ! w else zero_poly)"
    {
      fix w
      assume "?i \<noteq> w"
      hence "?exp w s = ?exp w t" by (simp add: nth_append, cases "w - ?i", auto)
    } note one = this
    {
      fix w
      have "\<exists> ts. (?exp w t = (if w < length ts then (map (eval_term I) ts) ! w else zero_poly))"
        by (rule exI[of _ ?t], simp only: map_append, simp)
      then obtain ts where id: "?exp w t = (if w < length ts then (map (eval_term I) ts) ! w else zero_poly)" by blast
      have "?exp w t \<ge>p zero_poly" 
        by (simp only: id, unfold poly_ge_def zero_poly_def, simp add: ge_refl, intro impI allI, force simp: eval_term_pos[unfolded poly_ge_def zero_poly_def, simplified])
    } note two = this 
    show "gt (eval_poly \<alpha> (eval_term I (Fun f ?s))) (eval_poly \<alpha> (eval_term I (Fun f ?t)))"
      by (simp, rule mono[unfolded poly_strict_mono_def poly_gt_def, THEN spec, THEN spec, THEN mp, THEN spec, THEN mp[OF _ pos]], rule conjI, (auto simp: one two)[1], simp add: nth_append st[unfolded inter_s_def poly_gt_def, simplified])
  qed
qed
end       

definition default_I :: "'a \<Rightarrow> nat \<Rightarrow> (nat,'a :: poly_carrier)poly"
where "default_I def n \<equiv> ([],def) # map (\<lambda> i. ([(i,1)],1)) [0 ..< n]"

lemma default_I_check_poly_weak_mono: assumes "ge def (0 :: 'a :: poly_carrier)" shows "check_poly_weak_mono (default_I def n)"
unfolding default_I_def check_poly_weak_mono_def
by (simp add: list_all_iff assms one_ge_zero)

lemma default_I_check_poly_strict_mono: assumes i: "i < n" shows "check_poly_strict_mono power_mono (default_I def n) i"
unfolding check_poly_strict_mono_def list_ex_iff default_I_def using i
by (auto simp: ge_refl check_monom_strict_mono_def)


lemma default_I_pos: assumes "ge def (0 :: 'a :: poly_carrier)" shows "default_I def n \<ge>p zero_poly"
using check_poly_weak_mono_pos[OF default_I_check_poly_weak_mono[OF assms]] .

lemma default_I_mono: assumes ge: "ge def (0 :: 'a :: poly_carrier)" shows "poly_weak_mono type (default_I def n)"
using check_poly_weak_mono[OF default_I_check_poly_weak_mono[OF assms]] .

locale ce_poly_order = poly_order +
  assumes default: "\<exists> n. \<forall> f m. m \<ge> n \<longrightarrow> I f m = default_I default m"

sublocale ce_poly_order \<subseteq> mono_ce_af_redpair inter_s inter_ns \<pi>
proof -
  have "\<exists> n. \<forall> m \<ge> n. \<forall> c d. ce_trs (c,d,m) \<subseteq> inter_ns \<inter> inter_s"
  proof -
    from default obtain n where default: "\<And> f m. m \<ge> n \<Longrightarrow> I f m = default_I default m" by auto
    show ?thesis
    proof(rule exI[of _ n], intro allI impI)
      fix m and c d :: 'b
      assume "n \<le> m"
      with default have default: "I c (Suc (Suc m)) = default_I default (Suc (Suc m))" by auto
      { 
        fix s t :: "('b,'c)term" and \<alpha> :: "('c,'a)assign"
        assume pos: "pos_assign \<alpha>"
        let ?t = "eval_poly \<alpha> (eval_term I t)"
        let ?s = "eval_poly \<alpha> (eval_term I s)"
        from eval_term_pos[of t] eval_term_pos[of s] have s: "?s \<ge> 0" and t: "?t \<ge> 0" using pos unfolding zero_poly_def poly_ge_def by auto
        have main: "eval_poly \<alpha> (eval_term I (Fun c (s # (t # replicate m (Fun d []))))) \<ge> default + ?s + ?t" (is "ge ?left ?dst")
        proof (simp add: default poly_subst, simp only: default_I_def upt_conv_Cons)
          let ?ass = "(\<lambda>v. eval_poly \<alpha> (if v < Suc (Suc m) then (eval_term I s # eval_term I t # replicate m (poly_subst (\<lambda>i. zero_poly) (I d 0))) ! v else zero_poly))"
          let ?ass2 = "(\<lambda>v. eval_poly \<alpha> (if v < Suc (Suc m) then (eval_term I s # eval_term I t # replicate m (eval_term I (Fun d []))) ! v else zero_poly))"
          let ?f = "(\<lambda>i :: nat. ([(i, 1 :: nat)], 1\<Colon>'a))"
          let ?is = "[Suc (Suc 0)..<Suc (Suc m)]"
          obtain ass iis where ass: "ass = ?ass" and iis: "iis = ?is" by auto
          hence "eval_poly ?ass (([],default) # map ?f (0 # Suc 0 # ?is)) = 
                 eval_poly ass (([],default) # map ?f (0 # Suc 0 # iis))" (is "?left = _") by simp
          also have "\<dots> = default + ?s + ?t + eval_poly ass (map ?f iis)" by (auto simp: field_simps ass)
          also have "\<dots> \<ge> default + ?s + ?t"
          proof -
            have "eval_poly ass (map ?f iis) \<ge> 0" 
            proof (rule check_poly_weak_mono_pos[unfolded poly_ge_def zero_poly_def, THEN spec[of _ ass], THEN mp, simplified])
              show "check_poly_weak_mono (map ?f iis)" unfolding check_poly_weak_mono_def list_all_iff 
                using one_ge_zero by auto
            next
              show "pos_assign ass" unfolding pos_assign_def
              proof
                fix i
                show "ass i \<ge> 0"
                proof (cases "i = 0", simp add: s ass, cases "i = 1", simp add: t ass, cases "i \<ge> Suc (Suc m)", simp add: ass zero_poly_def ge_refl)
                  assume "i \<noteq> 0" and "i \<noteq> 1" and "\<not> Suc (Suc m) \<le> i"
                  hence "i \<ge> 2" and "i < Suc (Suc m)" by auto
                  then obtain j where i: "i = Suc (Suc j)" and j: "j < m" by (cases i, simp, cases "i - 1", auto)
                  show "ass i \<ge> 0" unfolding i ass 
                  proof (simp add: j)
                    have "eval_poly \<alpha> (poly_subst (\<lambda> i. zero_poly) (I d 0)) = eval_poly \<alpha> (eval_term I (Fun d []))" (is "?left = _")
                      by simp
                    also have "\<dots> \<ge> 0" using eval_term_pos[unfolded poly_ge_def zero_poly_def, of "Fun d []"] pos by auto
                    finally show "?left \<ge> 0" .
                  qed
                qed
              qed
            qed
            from plus_left_mono[OF this, of "default + ?s + ?t"] show ?thesis by (auto simp: field_simps)
          qed
          finally show "?left \<ge> default + ?s + ?t" .
        qed
        from main have mains: "?left \<ge> ?dst" .
        also have "\<dots> \<ge> ?s + ?t" using plus_left_mono[OF default_ge_zero, of "?s + ?t"] by (auto simp: field_simps)
        finally have main: "?left \<ge> ?s + ?t" .
        from main have "?left \<ge> ?s + ?t" .
        also have "\<dots> \<ge> ?s" using plus_left_mono[OF t, of ?s] by (auto simp: field_simps)
        finally have sge: "?left \<ge> ?s" .
        from main have "?left \<ge> ?s + ?t" .
        also have "\<dots> \<ge> ?t" using plus_left_mono[OF s, of ?t] by (auto simp: field_simps)
        finally have tge: "?left \<ge> ?t" .
        from mains have "?left \<ge> ?dst" .
        also have "\<dots> \<ge> default + ?s" using plus_left_mono[OF t, of "default + ?s"] by (auto simp: field_simps)
        also have "gt \<dots> ?s" using plus_gt_left_mono[OF default_gt_zero, of ?s] by simp
        finally have sgt: "gt ?left ?s" .
        from mains have "?left \<ge> ?dst" .
        also have "\<dots> \<ge> default + ?t" using plus_left_mono[OF s, of "default + ?t"] by (auto simp: field_simps)
        also have "gt \<dots> ?t" using plus_gt_left_mono[OF default_gt_zero, of ?t] by simp
        finally have tgt: "gt ?left ?t" .
        from sge tge sgt tgt have "?left \<ge> ?s" and "?left \<ge> ?t" and "gt ?left ?s" and "gt ?left ?t" by blast+
      } note main = this
      show "ce_trs (c,d,m) \<subseteq> inter_ns \<inter> inter_s"
        unfolding ce_trs.simps inter_ns_def inter_s_def
        by (auto simp del: eval_term.simps simp: poly_ge_def poly_gt_def main)
    qed
  qed 
  hence one: "\<exists> n. \<forall> m \<ge> n. \<forall> c d. ce_trs (c,d,m) \<subseteq> inter_s" and
    two: "\<exists> n. \<forall> m \<ge> n. \<forall> c d. ce_trs (c,d,m) \<subseteq> inter_ns" by blast+
  show "mono_ce_af_redpair inter_s inter_ns \<pi>"
    by (unfold_locales, rule two, rule one)
qed

types ('f,'a)poly_inter_list = "(('f \<times> nat) \<times> (nat,'a)poly)list"

definition poly_inter_list_to_inter :: "'a :: poly_carrier \<Rightarrow> ('f,'a)poly_inter_list \<Rightarrow> ('f,'a)poly_inter"
where "poly_inter_list_to_inter def I \<equiv> \<lambda> f n. case lookup (f,n) I of None \<Rightarrow> default_I def n | Some p \<Rightarrow> p"

definition poly_inter_to_af :: "('f,'a)poly_inter \<Rightarrow> 'f af"
where "poly_inter_to_af I f n \<equiv> poly_var (I f n)"

definition check_poly_inter_list :: "('f,'a :: poly_carrier)poly_inter_list \<Rightarrow> ('f \<times> (nat,'a)poly) check"
where "check_poly_inter_list I \<equiv> check_all (\<lambda> (_,p). check_poly_weak_mono p) I <+? (\<lambda> ((f,n),p). (f,p))"

context poly_order_carrier
begin

lemma check_poly_inter_list: assumes check: "isOK(check_poly_inter_list I)"
  shows "check_poly_weak_mono (poly_inter_list_to_inter default I f n)"
proof -
  let ?I = "poly_inter_list_to_inter default I"
  show weak_mono: "check_poly_weak_mono (?I f n)"
    unfolding poly_inter_list_to_inter_def
  proof (cases "lookup (f,n) I", 
      simp add: default_I_check_poly_weak_mono[OF default_ge_zero], simp)
    fix p
    assume "lookup (f,n) I = Some p"
    from lookup_sound[OF this] check[unfolded check_poly_inter_list_def]
    show "check_poly_weak_mono p" by auto
  qed
qed

lemma ce_poly_order: assumes check: "isOK(check_poly_inter_list I)"
  shows "ce_poly_order default gt power_mono (poly_inter_list_to_inter default I) type (poly_inter_to_af (poly_inter_list_to_inter default I))"
proof -
  let ?I = "poly_inter_list_to_inter default I"
  from check_poly_inter_list[OF check] have main: "\<And> f n. check_poly_weak_mono (?I f n)" .
  show ?thesis
  proof (unfold_locales, rule check_poly_weak_mono_pos[OF main], rule check_poly_weak_mono[OF main])
    let ?m = "Suc (max_list (map (snd o fst) I))"
    show "\<exists> n. \<forall> f m. n \<le> m \<longrightarrow> ?I f m = default_I default m"
    proof (rule exI[of _ ?m], intro allI impI)
      fix f m 
      assume m: "?m \<le> m"
      show "?I f m = default_I default m" 
      proof (cases "lookup (f,m) I", simp add: poly_inter_list_to_inter_def)
        case (Some p)
        from lookup_sound[OF this] have "((f,m),p) \<in> set I" .
        hence "m \<in> set (map (snd o fst) I)" by force
        from max_list[OF this] m show ?thesis by auto
      qed
    qed
  next
    fix f n i
    assume "\<not> poly_inter_to_af (poly_inter_list_to_inter default I) f n i"
    thus "\<not> poly_var (poly_inter_list_to_inter default I f n) i"
      unfolding poly_inter_to_af_def .
  qed
qed
end

fun shows_monom :: "('v :: show)monom \<Rightarrow> string \<Rightarrow> string"
where "shows_monom [(x,p)] = (if p = 1 then shows x else shows x +@+ shows_string ''^'' +@+ shows p)"
   |  "shows_monom ((x,p) # m) = ((if p = 1 then shows x else shows x +@+ shows_string ''^'' +@+ shows p) +@+ shows_string ''*'' +@+ shows_monom m)"
   |  "shows_monom [] = shows_string ''1''"

fun shows_poly :: "('v :: show,'a :: {poly_carrier,show})poly \<Rightarrow> string \<Rightarrow> string"
where "shows_poly [] = shows_string ''0''"
  | "shows_poly ((m,c) # p) = ((if c = 1 then shows_monom m else shows c +@+ shows_string ''*'' +@+ shows_monom m) +@+ shows_string '' + '' +@+ shows_poly p)"

definition create_nlpoly_redpair :: "'a :: {show,poly_carrier} \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> bool \<Rightarrow> ('f,'a)poly_inter_list \<Rightarrow> ('f :: show,'v :: show)redpair_impl"
where "create_nlpoly_redpair def gt power_mono I = (let J = poly_inter_list_to_inter def I in 
    (check_poly_inter_list I <+? (\<lambda> (f,p). shows_string ''interpretation of '' +@+ shows_poly (poly_subst (\<lambda> n. poly_of (PVar (''x_'' @ (shows n [])))) p) +@+ shows_string '' of '' +@+ shows f +@+ shows_string '' invalid ''),
     (\<lambda> (s,t). let p = eval_term J s; q = eval_term J t in check (check_poly_gt gt p q) (shows_string ''could not ensure '' +@+ shows_poly p +@+ shows_string '' > '' +@+ shows_poly q)),
     (\<lambda> (s,t). let p = eval_term J s; q = eval_term J t in check (check_poly_ge p q) (shows_string ''could not ensure '' +@+ shows_poly p +@+ shows_string '' >= '' +@+ shows_poly q)),
     poly_inter_to_af J,
     check_all (\<lambda> ((f,n),p). list_all (\<lambda> i. check_poly_strict_mono power_mono p i) [0 ..< n]) I <+? (\<lambda> ((f,n),p). shows_string ''could not ensure monotonicty of '' +@+ shows_poly p +@+ 
 shows_string '' as interpretation of '' +@+ shows f),
     shows_string ''no representation for non-linear polynomial interpretations so far''))"

lemma poly_order_carrier_to_generic_redpair_impl: 
  fixes default :: "'a :: {show,poly_carrier}"
  and gt :: "'a \<Rightarrow> 'a \<Rightarrow> bool"
  and power_mono :: bool
  assumes "poly_order_carrier default gt power_mono"
  shows "generic_redpair_impl (create_nlpoly_redpair default gt power_mono)"
proof -
  from assms interpret poly_order_carrier default gt power_mono .
  show ?thesis
  proof
    fix I :: "('f :: show,'a)poly_inter_list" and s_list ns_list :: "('f,'v :: show)rule list" and type :: "'v"
    let ?rp = "create_nlpoly_redpair default gt power_mono I :: ('f,'v)redpair_impl"
    let ?af = "rp_af ?rp :: 'f af"
    assume valid: "isOK (rp_valid ?rp)" 
    assume s: "isOK (check_allm (rp_strict ?rp) s_list)" 
    assume ns: "isOK (check_allm (rp_non_strict ?rp) ns_list)"
    from valid have valid: "isOK(check_poly_inter_list I)" unfolding create_nlpoly_redpair_def by (auto simp: Let_def)
    let ?J = "poly_inter_list_to_inter default I :: ('f,'a)poly_inter"
    from ce_poly_order[OF valid] interpret ce_poly_order default gt power_mono ?J type ?af unfolding create_nlpoly_redpair_def by (simp add: Let_def) 
    have mrp: "mono_ce_af_redpair (inter_s :: ('f,'v)trs) inter_ns ?af" ..
    have rp: "ce_af_redpair (inter_s :: ('f,'v)trs) inter_ns ?af" ..
    show "\<exists> S NS. ce_af_redpair S NS (rp_af ?rp) \<and> set s_list \<subseteq> S \<and> set ns_list \<subseteq> NS \<and> (isOK(rp_mono ?rp)  \<longrightarrow> mono_ce_af_redpair S NS (rp_af ?rp) \<and> one_step_ctxt_closed S)"
    proof (intro exI allI conjI, rule rp)
      show "set s_list \<subseteq> inter_s" 
      proof
        fix s t
        assume "(s,t) \<in> set s_list"
        with s have "check_poly_gt gt (eval_term ?J s) (eval_term ?J t)" unfolding create_nlpoly_redpair_def
          by (auto simp: Let_def)
        from check_poly_gt[OF this] show "(s,t) \<in> inter_s" unfolding inter_s_def by simp
      qed
    next
      show "set ns_list \<subseteq> inter_ns" 
      proof
        fix s t
        assume "(s,t) \<in> set ns_list"
        with ns have "check_poly_ge (eval_term ?J s) (eval_term ?J t)" unfolding create_nlpoly_redpair_def
          by (auto simp: Let_def)
        from check_poly_ge[OF this] show "(s,t) \<in> inter_ns" unfolding inter_ns_def by simp
      qed
    next
      show " isOK (rp_mono ?rp) \<longrightarrow>
        mono_ce_af_redpair inter_s inter_ns ?af \<and>
        one_step_ctxt_closed inter_s"
      proof (intro impI conjI, simp add: mrp)
        assume mono: "isOK(rp_mono ?rp)"
        interpret mono_poly_order default gt power_mono ?J type ?af
        proof
          fix f :: 'f and  n i :: nat
          assume i: "i < n" 
          show "poly_strict_mono type (?J f n) i"          
          proof (rule check_poly_strict_mono[OF _ check_poly_inter_list[OF valid]])
            show "check_poly_strict_mono power_mono (?J f n) i"
              unfolding poly_inter_list_to_inter_def
            proof (cases "lookup (f,n) I", 
                simp add: default_I_check_poly_strict_mono[OF i], simp)
              fix p
              assume "lookup (f,n) I = Some p"
              from lookup_sound[OF this] mono[unfolded create_nlpoly_redpair_def, simplified Let_def, simplified, unfolded check_poly_inter_list_def list_all_iff] i
              show "check_poly_strict_mono power_mono p i" by force
            qed
          qed
        qed
        show "one_step_ctxt_closed inter_s" by (rule inter_s_mono)
      qed
    qed
  qed
qed

(* this did not work due to additional type constraint that 'a must be of sort show *)
(* sublocale poly_order_carrier \<subseteq> generic_redpair_impl "create_nlpoly_redpair default gt power_mono" *)


end
  